package org.netkernel.client.http.endpoint;

import org.apache.commons.httpclient.DefaultMethodRetryHandler;
import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.HttpMethodBase;
import org.apache.commons.httpclient.HttpState;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.MultiThreadedHttpConnectionManager;
import org.apache.commons.httpclient.NTCredentials;
import org.apache.commons.httpclient.UsernamePasswordCredentials;
import org.apache.commons.httpclient.auth.AuthScope;
import org.apache.commons.httpclient.methods.DeleteMethod;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.methods.HeadMethod;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.PatchMethod;
import org.apache.commons.httpclient.methods.PutMethod;
import org.apache.commons.httpclient.methods.RequestEntity;
import org.apache.commons.httpclient.util.DateParseException;
import org.apache.commons.httpclient.util.DateUtil;
import org.netkernel.client.http.representation.HttpCacheRepresentation;
import org.netkernel.client.http.representation.HttpStateRepresentation;
import org.netkernel.client.http.representation.IHttpClientConfig;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.representation.ByteArrayRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;

import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.util.Date;
import java.util.HashMap;


public class HttpClientAccessorOld extends StandardAccessorImpl
{	
	
	private MultiThreadedHttpConnectionManager mManager;
	private static HttpClientAccessorOld sInstance;
	
	private static final String NETKERNEL_USER_AGENT="1060 NetKernel HTTP Client 1.0.0  http://www.1060research.com";
	private static final String HTTP_HEADER_CONTENT_LENGTH="Content-Length";
	
	private static final String HTTP_ACCESSOR_STATUS_CODE_METADATA="HTTP_ACCESSOR_STATUS_CODE_METADATA";
	private static final String HTTP_ACCESSOR_RESPONSE_HEADERS_METADATA="HTTP_ACCESSOR_RESPONSE_HEADERS_METADATA";
	
	private static final String HTTP_METHOD_GET="httpGet";
	private static final int HTTP_METHOD_GET_INT=1;
	private static final String HTTP_METHOD_PUT="httpPut";
	private static final int HTTP_METHOD_PUT_INT=2;
	private static final String HTTP_METHOD_POST="httpPost";
	private static final int HTTP_METHOD_POST_INT=3;
	private static final String HTTP_METHOD_HEAD="httpHead";
	private static final int HTTP_METHOD_HEAD_INT=4;
	private static final String HTTP_METHOD_DELETE="httpDelete";
	private static final int HTTP_METHOD_DELETE_INT=5;
	private static final String HTTP_METHOD_PATCH="httpPatch";
	private static final int HTTP_METHOD_PATCH_INT=6;
	
	private static final String CONTENT_TYPE_HEADER="Content-Type";
	private static final String EXPIRES_HEADER="Expires";
	
	//private static final String HTTP_CONFIG_URI="res:/etc/HTTPClientConfig.xml";
	private static final String HTTP_CONFIG_URI_DEFAULT="res:/etc/DefaultHTTPClientConfig.xml";
	
	private static final HashMap mModes=new HashMap();
	{
		mModes.put(HTTP_METHOD_GET, new Integer(HTTP_METHOD_GET_INT));
		mModes.put(HTTP_METHOD_PUT, new Integer(HTTP_METHOD_PUT_INT));		
		mModes.put(HTTP_METHOD_POST, new Integer(HTTP_METHOD_POST_INT));
		mModes.put(HTTP_METHOD_HEAD, new Integer(HTTP_METHOD_HEAD_INT));
		mModes.put(HTTP_METHOD_DELETE, new Integer(HTTP_METHOD_DELETE_INT));
		mModes.put(HTTP_METHOD_PATCH, new Integer(HTTP_METHOD_PATCH_INT));
	}
	
	/** Creates a new instance of HTTPClientAccessor */
	public HttpClientAccessorOld()
	{	declareThreadSafe();
		System.getProperties().setProperty("httpclient.useragent", NETKERNEL_USER_AGENT);
		Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
		mManager=new MultiThreadedHttpConnectionManager();
		sInstance=this;
	}
	
	public static HttpClientAccessorOld getInstance()
	{	return sInstance;
	}
	
	@Override
	public void onDecommissionEndpoint() throws Exception
	{	mManager.shutdownAll();
	}

	@Override
	public void onSource(INKFRequestContext aContext) throws Exception
	{	Object result=null;
		HttpState state=null;
		Object stateref=null;
		int statusCode=-1;
		IHDSNode responseHeaders=null;
		long expires=0;
		String mimetype=null;
		boolean cache=aContext.getThisRequest().argumentExists("cache");
		
		//Get the configuration - first by push, then by pull
		IHttpClientConfig httpConfig;
		if(aContext.getThisRequest().argumentExists("config"))
		{	httpConfig=aContext.source("arg:config", IHttpClientConfig.class);
		}
		else
		{	httpConfig=aContext.source(HTTP_CONFIG_URI_DEFAULT, IHttpClientConfig.class);
		}
			
		mManager.getParams().setDefaultMaxConnectionsPerHost(httpConfig.getMaximumConnectionsPerHost());
		mManager.getParams().setMaxTotalConnections(httpConfig.getMaximumTotalConnections());
		
		//Retry handler
		DefaultMethodRetryHandler retryhandler = new DefaultMethodRetryHandler();
		retryhandler.setRequestSentRetryEnabled(false);
		retryhandler.setRetryCount(httpConfig.getNumberOfRetryAttempts());
		
		String url;
		if(aContext.getThisRequest().argumentExists("simple"))
		{	//Set the url from simple format
			url=aContext.getThisRequest().getArgumentValue("simple");
			GetMethod get=new GetMethod(url);
			get.setMethodRetryHandler(retryhandler);
			get.setFollowRedirects(httpConfig.followRedirects());
			try
			{	HttpClient client=getHttpClient(state,httpConfig);
	          	statusCode=client.executeMethod(get);
	          	if(statusCode>=400)
	          	{	throw new Exception("HTTP Request Error :"+statusCode);
	          	}
	          	result=getResponse(get);
	          	responseHeaders=getResponseHeaders(get.getResponseHeaders());
				mimetype=getContentType(get.getResponseHeader(CONTENT_TYPE_HEADER));
				expires=getExpires(get.getResponseHeader(EXPIRES_HEADER));
			}
			finally
			{	get.releaseConnection();
			}
			INKFResponse resp= aContext.createResponseFrom(result);
			resp.setHeader(HTTP_ACCESSOR_STATUS_CODE_METADATA, statusCode);
			resp.setHeader(HTTP_ACCESSOR_RESPONSE_HEADERS_METADATA,responseHeaders);
			resp.setMimeType(mimetype);
			if(0 < expires)
			{	resp.setExpiry(INKFResponse.EXPIRY_CONSTANT, expires);	//Expire at an absolute time
			}
			else
			{	resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
			}
		}
		else
		{	//Set the url
			url=aContext.getThisRequest().getArgumentValue("url");
			//Get state if there is any
			String stateuri=aContext.getThisRequest().getArgumentValue("state");
			
			if(stateuri!=null)
			{	if(!stateuri.startsWith(HttpStateAccessor.ACTIVE_HTTP_STATE))
				{ stateuri=aContext.source("arg:state",String.class);
				}
				HttpStateRepresentation sa=aContext.source(stateuri, HttpStateRepresentation.class);
				state=sa.getState();
				stateref=state.getCookies();
			}
			//User Specified Headers
			String headeruri=aContext.getThisRequest().getArgumentValue("headers");
			IHDSNode headers=null;
			if(headeruri!=null)
			{	headers=aContext.source("arg:headers", IHDSNode.class);
			}
					
			Integer mode=(Integer)mModes.get(aContext.getThisRequest().getArgumentValue("activeType"));
			switch (mode.intValue())
			{	case HTTP_METHOD_GET_INT:
					boolean remote=true;
					boolean shouldCache=false;
					HttpCacheRepresentation cacheRep=null;
					if(cache)
					{	//Source from cache
						INKFRequest req=aContext.createRequest(aContext.getThisRequest().getArgumentValue("cache"));
						req.addArgument("url", url);
						req.setRepresentationClass(HttpCacheRepresentation.class);
						cacheRep=(HttpCacheRepresentation)aContext.issueRequest(req);
						/*
						 * Should be test for expired
						 * 
						if(cacheRep!=null)
						{	remote=false;
							statusCode=HttpStatus.SC_OK;
							mimetype=cacheRep.getMimeType();
							result=cacheRep;
						}
						*/
					}
					if(remote)
					{	GetMethod get=new GetMethod(url);
						get.setMethodRetryHandler(retryhandler);
						get.setFollowRedirects(httpConfig.followRedirects());
						if(headers!=null)
						{	this.setRequestHeaders(get, headers);
						}
						if(cacheRep!=null)
						{	get.setRequestHeader("If-None-Match", cacheRep.getTag());
						}
						try
						{	HttpClient client=getHttpClient(state,httpConfig);
		                  	statusCode=client.executeMethod(get);
		                  	if(statusCode==HttpStatus.SC_NOT_MODIFIED)
		                  	{	result=cacheRep;
		                  	}
		                  	else
		                  	{	result=getResponse(get);
		                  		shouldCache=true;
		                  	}
		                  	responseHeaders=getResponseHeaders(get.getResponseHeaders());
							mimetype=getContentType(get.getResponseHeader(CONTENT_TYPE_HEADER));
							expires=getExpires(get.getResponseHeader(EXPIRES_HEADER));
						}
						finally
						{	get.releaseConnection();
						}
						//Sink to cache
						if(cache && shouldCache)
						{	INKFRequest req=aContext.createRequest(aContext.getThisRequest().getArgumentValue("cache"));
							req.addArgument("url", url);
							String etag=(String)responseHeaders.getFirstValue("ETag");
							req.addPrimaryArgument(new HttpCacheRepresentation(etag, mimetype, 0, (ByteArrayRepresentation)result ));
							req.setVerb(INKFRequestReadOnly.VERB_SINK);
							aContext.issueRequest(req);
						}
					}
				break;
				
				case HTTP_METHOD_POST_INT:
					PostMethod post=new PostMethod(url);
					post.setMethodRetryHandler(retryhandler);
					post.getParams().setParameter("http.protocol.expect-continue", Boolean.TRUE);
					if(headers!=null)
					{	this.setRequestHeaders(post, headers);
					}
					//post.setFollowRedirects(cfg.followRedirects());  Entity enclosing methods cannot follow redirects without user intervention
					if(aContext.getThisRequest().argumentExists("body")) 
					{	//Currently we're only supporting single body argument
						INKFResponseReadOnly<IReadableBinaryStreamRepresentation> resp=aContext.sourceForResponse("arg:body", IReadableBinaryStreamRepresentation.class);
						IReadableBinaryStreamRepresentation rep=resp.getRepresentation();
						post.setRequestEntity(new ReadableBinaryRequestEntity(rep, resp.getMimeType()));
					}
					else if(aContext.getThisRequest().argumentExists("nvp"))
					{	IHDSNode nvpa=(IHDSNode)aContext.source("arg:nvp",IHDSNode.class);
						for(IHDSNode n: nvpa.getChildren())
						{	post.addParameter(n.getName(), (String)n.getValue());
						}
					}
					try
					{	HttpClient client=getHttpClient(state,httpConfig);
	                  	statusCode=client.executeMethod(post);
						result=getResponse(post);	
						responseHeaders=getResponseHeaders(post.getResponseHeaders());
						mimetype=getContentType(post.getResponseHeader(CONTENT_TYPE_HEADER));
						expires=getExpires(post.getResponseHeader(EXPIRES_HEADER));
					}
					finally
					{	post.releaseConnection();
					}
				break;
				case HTTP_METHOD_PUT_INT:
				{
					PutMethod put=new PutMethod(url);
					put.setMethodRetryHandler(retryhandler);
					put.getParams().setParameter("http.protocol.expect-continue", Boolean.TRUE);
					if(headers!=null)
					{	this.setRequestHeaders(put, headers);
					}
					//put.setFollowRedirects(cfg.followRedirects());  Entity enclosing methods cannot follow redirects without user intervention
					INKFResponseReadOnly resp=aContext.sourceForResponse("arg:body", IReadableBinaryStreamRepresentation.class);
					IReadableBinaryStreamRepresentation rep=(IReadableBinaryStreamRepresentation)resp.getRepresentation();
					put.setRequestEntity(new ReadableBinaryRequestEntity(rep, resp.getMimeType()));
					try
					{	HttpClient client=getHttpClient(state,httpConfig);
	                 	statusCode=client.executeMethod(put);
	                 	result=getResponse(put);	
						responseHeaders=getResponseHeaders(put.getResponseHeaders());
						mimetype=getContentType(put.getResponseHeader(CONTENT_TYPE_HEADER));
						expires=getExpires(put.getResponseHeader(EXPIRES_HEADER));
					}
					finally
					{	put.releaseConnection();						
					}
				}
				break;
				case HTTP_METHOD_PATCH_INT:
				{
					PatchMethod patch=new PatchMethod(url);
					patch.setMethodRetryHandler(retryhandler);
					patch.getParams().setParameter("http.protocol.expect-continue", Boolean.TRUE);
					if(headers!=null)
					{	this.setRequestHeaders(patch, headers);
					}
					//put.setFollowRedirects(cfg.followRedirects());  Entity enclosing methods cannot follow redirects without user intervention
					INKFResponseReadOnly resp=aContext.sourceForResponse("arg:body", IReadableBinaryStreamRepresentation.class);
					IReadableBinaryStreamRepresentation rep=(IReadableBinaryStreamRepresentation)resp.getRepresentation();
					patch.setRequestEntity(new ReadableBinaryRequestEntity(rep, resp.getMimeType()));
					try
					{	HttpClient client=getHttpClient(state,httpConfig);
	                 	statusCode=client.executeMethod(patch);
	                 	result=getResponse(patch);	
						responseHeaders=getResponseHeaders(patch.getResponseHeaders());
						mimetype=getContentType(patch.getResponseHeader(CONTENT_TYPE_HEADER));
						expires=getExpires(patch.getResponseHeader(EXPIRES_HEADER));
					}
					finally
					{	patch.releaseConnection();						
					}
				}
				break;
				case HTTP_METHOD_HEAD_INT:
					HeadMethod head=new HeadMethod(url);
					head.setMethodRetryHandler(retryhandler);
					head.setFollowRedirects(httpConfig.followRedirects());
					if(headers!=null)
					{	this.setRequestHeaders(head, headers);
					}
					try
					{	HttpClient client=getHttpClient(state,httpConfig);
	                  	statusCode=client.executeMethod(head);
	                  	responseHeaders=getResponseHeaders(head.getResponseHeaders());
	                  	result=responseHeaders;
						mimetype=getContentType(head.getResponseHeader(CONTENT_TYPE_HEADER));
						expires=getExpires(head.getResponseHeader(EXPIRES_HEADER));
					}
					finally
					{	head.releaseConnection();						
					}
				break;
				case HTTP_METHOD_DELETE_INT:
					DeleteMethod delete=new DeleteMethod(url);
					delete.setMethodRetryHandler(retryhandler);
					//delete.setFollowRedirects(cfg.followRedirects());
					if(headers!=null)
					{	this.setRequestHeaders(delete, headers);
					}
					try
					{	HttpClient client=getHttpClient(state,httpConfig);
	                  	statusCode=client.executeMethod(delete);
	                  	result=getResponse(delete);
	                  	responseHeaders=getResponseHeaders(delete.getResponseHeaders());
						mimetype=getContentType(delete.getResponseHeader(CONTENT_TYPE_HEADER));
						expires=getExpires(delete.getResponseHeader(EXPIRES_HEADER));
					}
					finally
					{	delete.releaseConnection();						
					}
					
				break;
			}
			
			//Sink updated state
			if(state!=null && !state.getCookies().equals(stateref))
			{	HttpStateRepresentation sa=new HttpStateRepresentation(state);
				aContext.sink(stateuri, sa);
			}
			
			INKFResponse resp= aContext.createResponseFrom(result);
			resp.setHeader(HTTP_ACCESSOR_STATUS_CODE_METADATA, statusCode);
			resp.setHeader(HTTP_ACCESSOR_RESPONSE_HEADERS_METADATA,responseHeaders);
			resp.setMimeType(mimetype);
			//resp.setNoCache();
			if(expires>0)
			{	resp.setExpiry(INKFResponse.EXPIRY_CONSTANT, expires);	//Expire at an absolute time
			}
			else
			{	resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
			}

		}
	}
	
	private IHDSNode getResponseHeaders(Header[] aHeaders)
	{	HDSBuilder b=new HDSBuilder();
		for(Header h: aHeaders)
		{	b.addNode(h.getName(),h.getValue());
		}
		return b.getRoot();
	}
		
	private void setRequestHeaders(HttpMethod method, IHDSNode headers)
	{	for(IHDSNode n: headers.getChildren())
		{	String value=(String)n.getValue();
			method.setRequestHeader(n.getName(),value);
		}
	}
	
	private ByteArrayRepresentation getResponse(HttpMethodBase base) throws Exception
	{	return new ByteArrayRepresentation(getResponseInner(base).toByteArray());
	}
	
	private ByteArrayOutputStream getResponseInner(HttpMethodBase base) throws Exception
	{	int size;
		try
		{	Header h=base.getRequestHeader(HTTP_HEADER_CONTENT_LENGTH);
			size=Integer.parseInt(h.getValue());
		}
		catch(Exception e)
		{	//default to 1024 if no length specified
			size=1024;
		}
		ByteArrayOutputStream baos=new ByteArrayOutputStream(size);
		org.netkernel.layer0.util.Utils.pipe(
			base.getResponseBodyAsStream(),
			new BufferedOutputStream(baos)
		);
		return baos;
	}

	private String getContentType(Header h)
	{	String result;
		if(h!=null)
		{	String val=h.getValue();
			int i=val.indexOf(';');
			if(i>0)	result=val.substring(0,i);
			else result=val;
		}
		else result="content/unknown";
		return result;
	}
	
	private long getExpires(Header h)
	{	long result=0;
		try
		{	if(h!=null)
			{	Date d2=DateUtil.parseDate( h.getValue());
				result=d2.getTime();
			}
		}
		catch(DateParseException e){/*Must be expired*/}
		return result;
	}
	
	private HttpClient getHttpClient(HttpState state, IHttpClientConfig cfg)
	{	HttpClient client=new HttpClient(mManager);
		
		client.setConnectionTimeout(cfg.getConnectTimeout());
		client.setTimeout(cfg.getTimeout());
		String proxyHost=cfg.getProxyHost();
		if(proxyHost!=null)
		{	client.getHostConfiguration().setProxy( proxyHost, cfg.getProxyPort() );
		}
		String user=cfg.getProxyUser();
		String password=cfg.getProxyPassword();
		if (cfg.getProxyNTLMDomain()!=null)
		{	NTCredentials cred = new NTCredentials(user, password, cfg.getProxyNTLMHost(), cfg.getProxyNTLMDomain());
			state.setProxyCredentials(AuthScope.ANY, cred);
		}
		else if (user!=null && password!=null)
		{	
			if (state==null) state=new HttpState();
			UsernamePasswordCredentials cred=new UsernamePasswordCredentials(user,password);
			state.setProxyCredentials(AuthScope.ANY, cred);
		}
		
		if(state!=null) client.setState(state);
		
		return client;
	}
	
	private class ReadableBinaryRequestEntity implements RequestEntity
	{	
		private IReadableBinaryStreamRepresentation mRBS;
		private String mType;
		
		public ReadableBinaryRequestEntity(IReadableBinaryStreamRepresentation aRBS, String aType)
		{	mRBS=aRBS;
			mType=aType;
		}
		
		public long getContentLength()
		{	return mRBS.getContentLength();
		}
		
		public String getContentType()
		{	return mType;
		}
		
		public boolean isRepeatable()
		{	return true;
		}
		
		public void writeRequest(java.io.OutputStream outputStream) throws java.io.IOException
		{	mRBS.write(outputStream);
		}
	}
}
